# Generated by Django 3.2.13 on 2022-06-17 14:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("google_sheets", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="filewatchrenewalattempt",
            name="sheet_type",
            field=models.CharField(db_index=True, max_length=30),
        ),
    ]
